from datetime import datetime
from typing import Optional

from cpl_core.database import TableABC
from cpl_core.dependency_injection import ServiceProviderABC
from cpl_discord.service import DiscordBotServiceABC
from cpl_query.extension import List

from bot_data.model.level import Level
from bot_data.model.server import Server


class User(TableABC):
    def __init__(
        self,
        dc_id: int,
        xp: int,
        message_count: int,
        reaction_count: int,
        server: Optional[Server],
        created_at: datetime = None,
        modified_at: datetime = None,
        id=0,
    ):
        self._user_id = id
        self._discord_id = dc_id
        self._xp = xp
        self._message_count = message_count
        self._reaction_count = reaction_count
        self._server = server

        TableABC.__init__(self)
        self._created_at = created_at if created_at is not None else self._created_at
        self._modified_at = modified_at if modified_at is not None else self._modified_at

    @property
    def id(self) -> int:
        return self._user_id

    @property
    def discord_id(self) -> int:
        return self._discord_id

    @property
    @ServiceProviderABC.inject
    def name(self, bot: DiscordBotServiceABC) -> str:
        guild = bot.get_guild(self.server.discord_id)
        user = guild.get_member(self.discord_id)
        return None if user is None else user.name

    @property
    @ServiceProviderABC.inject
    def icon_url(self, bot: DiscordBotServiceABC) -> str:
        guild = bot.get_guild(self.server.discord_id)
        user = guild.get_member(self.discord_id)
        return None if user is None else user.display_icon

    @property
    def xp(self) -> int:
        return self._xp

    @xp.setter
    def xp(self, value: int):
        self._xp = value

    @property
    def message_count(self) -> int:
        return self._message_count

    @message_count.setter
    def message_count(self, value: int):
        self._message_count = value

    @property
    def reaction_count(self) -> int:
        return self._reaction_count

    @reaction_count.setter
    def reaction_count(self, value: int):
        self._reaction_count = value

    @property
    @ServiceProviderABC.inject
    def ontime(self, services: ServiceProviderABC) -> float:
        from bot_core.abc.client_utils_abc import ClientUtilsABC

        client_utils: ClientUtilsABC = services.get_service(ClientUtilsABC)
        return client_utils.get_ontime_for_user(self)

    @property
    @ServiceProviderABC.inject
    def level(self, services: ServiceProviderABC) -> Level:
        from modules.level.service.level_service import LevelService

        levels: LevelService = services.get_service(LevelService)
        return levels.get_level(self)

    @property
    def server(self) -> Optional[Server]:
        return self._server

    @property
    @ServiceProviderABC.inject
    def left_server(
        self,
        services: ServiceProviderABC,
    ) -> bool:
        from bot_data.abc.user_joined_server_repository_abc import UserJoinedServerRepositoryABC

        ujs: UserJoinedServerRepositoryABC = services.get_service(UserJoinedServerRepositoryABC)
        return ujs.find_active_user_joined_server_by_user_id(self.id) is None

    @property
    @ServiceProviderABC.inject
    def game_idents(
        self,
        services: ServiceProviderABC,
    ) -> List["UserGameIdent"]:
        from bot_data.abc.user_game_ident_repository_abc import UserGameIdentRepositoryABC

        game_idents_repo: UserGameIdentRepositoryABC = services.get_service(UserGameIdentRepositoryABC)
        return game_idents_repo.get_user_game_idents_by_user_id(self.id)

    @staticmethod
    def get_select_all_string() -> str:
        return str(
            f"""
            SELECT * FROM `Users`;
        """
        )

    @staticmethod
    def get_select_by_id_string(id: int) -> str:
        return str(
            f"""
            SELECT * FROM `Users`
            WHERE `UserId` = {id};
        """
        )

    @staticmethod
    def get_select_by_server_id_string(server_id: int) -> str:
        return str(
            f"""
            SELECT * FROM `Users`
            WHERE `ServerId` = {server_id};
        """
        )

    @staticmethod
    def get_select_by_discord_id_string(id: int) -> str:
        return str(
            f"""
            SELECT * FROM `Users`
            WHERE `DiscordId` = {id};
        """
        )

    @staticmethod
    def get_select_by_discord_id_and_server_id_string(dc_id: int, s_id: int) -> str:
        return str(
            f"""
            SELECT * FROM `Users`
            WHERE `DiscordId` = {dc_id}
            AND `ServerId` = {s_id};
        """
        )

    @property
    def insert_string(self) -> str:
        return str(
            f"""
            INSERT INTO `Users` (
                `DiscordId`, `XP`, `MessageCount`, `ReactionCount`, `ServerId`
            ) VALUES (
                {self._discord_id},
                {self._xp},
                {self._message_count},
                {self._reaction_count},
                {self._server.id}
            );
        """
        )

    @property
    def udpate_string(self) -> str:
        return str(
            f"""
            UPDATE `Users`
            SET `XP` = {self._xp},
            `MessageCount` = {self._message_count},
            `ReactionCount` = {self._reaction_count}
            WHERE `UserId` = {self._user_id};
        """
        )

    @property
    def delete_string(self) -> str:
        return str(
            f"""
            DELETE FROM `Users`
            WHERE `UserId` = {self._user_id};
        """
        )
